import 'package:path/path.dart';
import 'package:sqflite/sqflite.dart';
import 'package:sqlentity/base-entity/entity.dart';
import 'package:sqlentity/database/database-config.dart';

///The DataBaseConfig looks for the database configuration provided by the developer
class DataBaseBase {
  bool _openOne = true;
  Database _database;
  DataBaseConfig _databaseConfig = DataBaseConfig();
  static DataBaseBase _DATABASE_BASE;

  ///configured instance database
  factory DataBaseBase() {
    if (_DATABASE_BASE == null) {
      _DATABASE_BASE = DataBaseBase._getInstance();
    }
    return _DATABASE_BASE;
  }

  DataBaseBase._getInstance();

  ///database instantiated and configured
  Database get database => _database;

  ///open database
  open() async {
    if (!(_database != null && _database.isOpen)) {
      if (_openOne) {
        _database = await _createDataBase();
        _openOne = false;
      } else {
        _database = await _openDataBase();
      }
    }
  }

  ///close database
  Future close() async {
    if ((_database != null && _database.isOpen)) {
      await _database.close();
      _database = null;
    }
  }

  Future<Database> _openDataBase() async {
    return await openDatabase(
        join(await getDatabasesPath(), _databaseConfig.database_name));
  }

  ///create or update database
  Future<Database> _createDataBase() async {
    return await openDatabase(
        join(await getDatabasesPath(), _databaseConfig.database_name),
        version: _databaseConfig.database_version,
        singleInstance: true, onCreate: (Database db, int version) async {
      await _createTableBase(db);
    }, onUpgrade: (Database db, int oldVersion, int newVersion) async {
      await _createTableBase(db);
      for (var entity in _databaseConfig.entitys) {
        await _alterTableBase(db, entity, newVersion);
        await _deleteTableBase(db, entity, newVersion);
      }
    });
  }

  Future<void> _createTableBase(Database db) async {
    for (var entity in _databaseConfig.entitys) {
      var sql = "CREATE TABLE IF NOT EXISTS ${entity.table} (";

      entity.columncreate.forEach((column) {
        if (column.column == entity.column[entity.column.length - 1].column)
          sql += "${column.column} ${column.type}";
        else
          sql += "${column.column} ${column.type},";
      });

      sql += ")";

      try {
        await db.transaction((txn) async {
          var batch = txn.batch();
          batch.execute(sql);
          await batch.commit();
        });
      } catch (e) {
        await db.close();
      }
    }
  }

  Future<void> _alterTableBase(
      Database db, Entity entity, int newVersion) async {
    for (var columnalter in entity.columnalter) {
      if (columnalter.value == newVersion) {
        try {
          var sql = "CREATE TABLE IF NOT EXISTS TEMPORARY (";

          entity.columncreate.forEach((column) {
            if (column.column == columnalter.column) {
              if (columnalter.column ==
                  entity.column[entity.column.length - 1].column)
                sql += "${columnalter.column} ${columnalter.type}";
              else
                sql += "${columnalter.column} ${columnalter.type},";
            } else {
              if (column.column ==
                  entity.column[entity.column.length - 1].column)
                sql += "${column.column} ${column.type}";
              else
                sql += "${column.column} ${column.type},";
            }
          });
          sql += ")";

          await db.transaction((txn) async {
            var batch = txn.batch();
            batch.execute(sql);
            await batch.commit();
          });

          var sqlInsert = 'INSERT INTO TEMPORARY SELECT ';

          entity.columncreate.forEach((column) {
            if (column.column == columnalter.column) {
              if (columnalter.column ==
                  entity.column[entity.column.length - 1].column)
                sqlInsert += "${columnalter.column}";
              else
                sqlInsert += "${columnalter.column},";
            } else {
              if (column.column ==
                  entity.column[entity.column.length - 1].column)
                sqlInsert += "${column.column}";
              else
                sqlInsert += "${column.column},";
            }
          });

          await db.transaction((txn) async {
            var batch = txn.batch();
            sqlInsert += " FROM ${entity.table}";
            batch.execute(sqlInsert);
            var sqlDROP = "DROP TABLE ${entity.table}";
            batch.execute(sqlDROP);
            var sqlALTER = "ALTER TABLE TEMPORARY RENAME TO ${entity.table}";
            batch.execute(sqlALTER);
            await batch.commit();
          });
        } catch (e) {
          db.close();
        }
      }
    }
  }

  Future<void> _deleteTableBase(
      Database db, Entity entity, int newVersion) async {
    for (var columndelete in entity.columndelete) {
      if (columndelete.value == newVersion) {
        try {
          var sql = "CREATE TABLE IF NOT EXISTS TEMPORARY (";

          entity.columncreate.forEach((column) {
            if (column.column != columndelete.column) {
              if (column.column ==
                  entity.column[entity.column.length - 1].column)
                sql += "${column.column} ${column.type}";
              else
                sql += "${column.column} ${column.type},";
            }
          });

          sql += ")";

          await db.transaction((txn) async {
            var batch = txn.batch();
            batch.execute(sql);
            await batch.commit();
          });

          var sqlInsert = 'INSERT INTO TEMPORARY SELECT ';

          entity.columncreate.forEach((column) {
            if (column.column != columndelete.column) {
              if (column.column ==
                  entity.column[entity.column.length - 1].column)
                sqlInsert += "${column.column}";
              else
                sqlInsert += "${column.column},";
            }
          });

          await db.transaction((txn) async {
            var batch = txn.batch();
            sqlInsert += " FROM ${entity.table}";
            batch.execute(sqlInsert);
            var sqlDROP = "DROP TABLE ${entity.table}";
            batch.execute(sqlDROP);
            var sqlALTER = "ALTER TABLE TEMPORARY RENAME TO ${entity.table}";
            batch.execute(sqlALTER);
            batch.execute(sql);
            await batch.commit();
          });
        } catch (e) {
          db.close();
        }
      }
    }
  }
}
